#!/usr/bin/env python
"""
plot.py -- simple plotting utilities for quot

"""
import sys
import os 

# Numeric
import numpy as np 
from scipy import ndimage as ndi 

# Dataframes, for handling trajectories
import pandas as pd 

# Core matplotlib library
import matplotlib.pyplot as plt 

# Color schema
from matplotlib import cm
from matplotlib import colors as mpl_colors 

# 3D plots
from mpl_toolkits.mplot3d import Axes3D

########################
## PLOTTING UTILITIES ##
########################

def hex_cmap(cmap, n_colors):
    """
    Generate a matplotlib colormap as a list of hex colors
    indices.

    args
    ----
        cmap        :   str, name of a matplotlib cmap (for 
                        instance, "viridis")
        n_colors    :   int

    returns
    -------
        list of str, hex color codes

    """
    C = cm.get_cmap(cmap, n_colors)
    return [mpl_colors.rgb2hex(j[:3]) for j in C.colors]

def kill_ticks(axes, spines=True):
    """
    Remove the y and x ticks from a plot.

    args
    ----
        axes        :   matplotlib.axes.Axes
        spines      :   bool, also remove the spines

    returns
    -------
        None

    """
    axes.set_xticks([])
    axes.set_yticks([])
    if spines:
        for s in ['top', 'bottom', 'left', 'right']:
            axes.spines[s].set_visible(False)

def wrapup(out_png, dpi=600):
    """
    Save a figure to a PNG.

    args
    ----
        out_png         :   str, save filename
        dpi             :   int, resolution

    """
    plt.tight_layout()
    plt.savefig(out_png, dpi=dpi)
    plt.close()
    if sys.platform == "darwin":
        os.system("open {}".format(out_png))

####################
## IMAGE PLOTTING ##
####################

def imshow(*imgs, vmax_mod=1.0, plot=True):
    """
    Show a variable number of images side-by-side in a 
    temporary window.

    args
    ----
        imgs        :   2D ndarrays
        vmax_mod    :   float, modifier for the
                        white point as a fraction of 
                        max intensity

    """
    n = len(imgs)
    if n == 1:
        fig, ax = plt.subplots(figsize=(3, 3))
        ax.imshow(imgs[0], cmap='gray', vmax=vmax_mod*imgs[0].max())
    else:
        fig, ax = plt.subplots(1, n, figsize=(3*n, 3))
        for j in range(n):
            ax[j].imshow(imgs[j], cmap='gray', vmax=vmax_mod*imgs[j].max())

    if plot:
        plt.show(); plt.close()
    else:
        return fig, ax 

def max_int_proj(axes, nd2_path, vmax_perc=99, vmin=None, cmap="gray",
    pixel_size_um=0.16, scalebar=False):
    """
    Make a maximum intensity projection of a temporal image sequence
    in a Nikon ND2 file.

    args
    ---
        axes        :   matplotlib.axes.Axes
        nd2_path    :   str, path to a target ND2 file
        vmax_perc   :   float, percentile of the intensity histogram
                        to use for the upper contrast bound
        vmin        :   float
        cmap        :   str
        pixel_size_um:  float, size of pixels in um
        scalebar    :   bool, make a scalebar

    returns
    -------
        None; plots directly to *axes*

    """
    from nd2reader import ND2Reader 
    
    reader = ND2Reader(nd2_path)
    mip = reader.get_frame(0).astype(np.float64)
    for frame_idx in range(1, reader.metadata["total_images_per_channel"]):
        try:
            mip = np.maximum(mip, reader.get_frame(frame_idx))
        except:
            continue
    if vmin is None:
        vmin = mip.min()
    axes.imshow(mip, cmap=cmap, vmin=vmin,
        vmax=np.percentile(mip, vmax_perc))
    kill_ticks(axes)
    if scalebar:
        try:
            from matplotlib_scalebar.scalebar import ScaleBar
            sb = ScaleBar(pixel_size_um, "um", frameon=False, 
                color="w", location="lower left")
            axes.add_artist(sb)
        except ModuleNotFoundError:
            print("quot.plot.max_int_proj: must have matplotlib_scalebar " \
                "installed in order to use scalebars")

def imlocdensity(axes, tracks, ymax=None, xmax=None, bin_size=0.1,
    kernel=3.0, vmax_perc=99, cmap="gray", pixel_size_um=0.16,
    scalebar=False):
    """
    Given a set of localizations from a single FOV, plot localization
    density.

    args
    ----
        axes        :   matplotlib.axes.Axes
        tracks      :   pandas.DataFrame
        ymax        :   int, the height of the FOV in pixels
        xmax        :   int, the width of the FOV in pixels
        bin_size    :   float, the size of the histogram bins in 
                        terms of pixels
        kernel      :   float, the size of the Gaussian kernel 
                        used for density estimation
        vmax_perc   :   float, the percentile of the density  
                        histogram used to define the upper contrast
                        threshold
        cmap        :   str
        pixel_size_um   :   float, size of pixels in um
        scalebar    :   bool, make a scalebar. Requires the 
                        matplotlib_scalebar package.

    returns
    -------
        None; plots directly to *axes*

    """
    # If ymax and xmax are not provided, set them to the maximum
    # y and x coordinates observed in the data
    if ymax is None:
        ymax = int(tracks['y'].max()) + 1
    if xmax is None:
        xmax = int(tracks['x'].max()) + 1

    # Determine histogram binning scheme
    ybins = np.arange(0, ymax, bin_size)
    xbins = np.arange(0, xmax, bin_size)

    # Accumulate localizations into a histogram
    density = np.histogram2d(tracks['y'], tracks['x'],
        bins=(ybins, xbins))[0].astype(np.float64)

    # KDE
    density = ndi.gaussian_filter(density, kernel)

    # Plot the result
    if density.sum() == 0:
        s = axes.imshow(density, vmin=0, vmax=1, cmap=cmap)
    else:
        s = axes.imshow(density, vmin=0,
            vmax=np.percentile(density, vmax_perc), 
            cmap=cmap)

    # Kill ticks
    kill_ticks(axes)

    # Scalebar, if desired
    if scalebar:
        try:
            from matplotlib_scalebar.scalebar import ScaleBar
            sb = ScaleBar(pixel_size_um*bin_size, "um", 
                frameon=False, color="w", location="lower left")
            axes.add_artist(sb)
        except ModuleNotFoundError:
            print("WARNING: module matplotlib_scalebar must be installed " \
                "in order to use scalebars")

###############################
## PSF PLOTTING and 3D PLOTS ##
###############################

def wireframe_overlay(img, model, plot=True):
    """
    Make a overlay of two 2-dimensional functions.

    args
    ----
        img, model      :   2D ndarray (YX), with the
                            same shape 

    """
    assert img.shape == model.shape 
    fig = plt.figure(figsize=(4,4))
    ax = fig.add_subplot(projection='3d')
    Y, X = np.indices(img.shape)
    ax.plot_wireframe(X, Y, img, color='k')
    ax.plot_wireframe(X, Y, model, color='r')

    if plot:
        plt.show(); plt.close()
    else:
        return fig, ax 

################################
## TRAJECTORY ATTRIBUTE PLOTS ##
################################

def locs_per_frame(axes, tracks, n_frames=None, kernel=5,
    fontsize=10, title=None):
    """
    Plot the number of localizations per frame.

    args
    ----
        axes        :   matplotlib.axes.Axes
        tracks      :   pandas.DataFrame
        n_frames    :   int, the number of frames in the
                        movie. If *None*, defaults to the 
                        maximum frame in *tracks*
        kernel      :   float, size of uniform kernel used
                        for smoothing

    returns
    -------
        None

    """
    if n_frames is None:
        n_frames = int(tracks["frame"].max()) + 1
    frame_indices = np.arange(0, n_frames+1)
    H = np.histogram(
        tracks["frame"],
        bins=frame_indices,
    )[0].astype(np.float64)
    if kernel > 0:
        H = ndi.uniform_filter(H, kernel)
    axes.plot(frame_indices[:-1], H, color="k")
    axes.set_xlabel("Frame", fontsize=fontsize)
    axes.set_ylabel("Spots per frame", fontsize=fontsize)
    if not title is None:
        axes.set_title(title, fontsize=fontsize)

############################
## JUMP LENGTH HISTOGRAMS ##
############################

def plot_jump_pdfs(axes, tracks, n_frames=4, pixel_size_um=0.16, bin_size=0.02,
    norm=True, frame_interval=0.00748, fontsize=10, cmap="viridis",
    start_frame=0, max_jumps_per_track=4, use_entire_track=False, **kwargs):
    """
    Plot the empirical jump length probability density function.

    args
    ----
        axes            :   matplotlib.axes.Axes
        tracks          :   pandas.DataFrame
        n_frames        :   int, the maximum number of frame 
                            intervals to consider
        pixel_size_um   :   float, size of pixels in um
        bin_size        :   float, size of the bins to use in 
                            the plotted histogram
        norm            :   bool, normalize to a PDF
        frame_interval  :   float, frame interval in seconds
        fontsize        :   int
        cmap            :   str
        start_frame     :   int, disregard jumps before this frame
        max_jumps_per_track     :   int, the maximum number of jumps
                                    to consider from any one track
        use_entire_track:   bool, use all jumps from every track
        kwargs          :   to rad_disp_histogram_2d

    """
    if start_frame > 0:
        tracks = tracks[tracks["frame"] >= start_frame]

    # Make the complete jump length histogram
    H, bin_edges = rad_disp_histogram_2d(tracks, n_frames=n_frames,
        pixel_size_um=pixel_size_um, use_entire_track=use_entire_track,
        max_jumps_per_track=max_jumps_per_track, **kwargs)

    # Aggregate histogram bins
    factor = int(bin_size // 0.001)
    H, bin_edges = coarsen_histogram(H, bin_edges, factor)
    bin_size = bin_edges[1] - bin_edges[0]
    bin_c = bin_edges[:-1] + bin_size * 0.5

    if norm:
        H = H.astype(np.float64)
        for frame_idx in range(n_frames):
            if H[frame_idx, :].sum() > 0:
                H[frame_idx, :] = H[frame_idx, :] / H[frame_idx, :].sum()

    # Plot
    colors = hex_cmap(cmap, n_frames+2)
    for frame_idx in range(n_frames):
        axes.plot(bin_c, H[frame_idx, :], color=colors[frame_idx],
            label="$\Delta t = $ %.4f sec" % (frame_interval * (frame_idx+1)))
    axes.set_xlabel("2D radial jump ($\mu$m)", fontsize=fontsize)
    axes.set_ylabel("Frequency")
    axes.legend(frameon=False, loc="upper right", prop={'size': 8})

def plot_jump_cdfs(axes, tracks, n_frames=4, pixel_size_um=0.16,
    frame_interval=0.00748, fontsize=10, cmap="viridis", 
    start_frame=0, max_jumps_per_track=4, use_entire_track=False, 
    linewidth=1, plot_max_r=None, **kwargs):
    """
    Plot the empirical jump length probability cumulative
    distribution function.

    args
    ----
        axes            :   matplotlib.axes.Axes
        tracks          :   pandas.DataFrame
        n_frames        :   int, the maximum number of frame 
                            intervals to consider
        pixel_size_um   :   float, size of pixels in um
        bin_size        :   float, size of the bins to use in 
                            the plotted histogram
        frame_interval  :   float, frame interval in seconds
        fontsize        :   int
        cmap            :   str
        start_frame     :   int, disregard jumps before this frame
        max_jumps_per_track     :   int, the maximum number of jumps
                                    to consider from any one track
        use_entire_track:   bool, use all jumps from every track
        linewidth       :   int, width of the lines
        plot_max_r      :   float, the maximum jump length to show
                            in the plot
        kwargs          :   to rad_disp_histogram_2d

    """
    if start_frame > 0:
        tracks = tracks[tracks['frame'] >= start_frame]
    H, bin_edges = rad_disp_histogram_2d(tracks, n_frames=n_frames,
        pixel_size_um=pixel_size_um, max_jumps_per_track=max_jumps_per_track,
        use_entire_track=use_entire_track, **kwargs)
    H = H.astype(np.float64)
    cdfs = np.cumsum(H, axis=1)
    for i in range(n_frames):
        cdfs[i,:] = cdfs[i,:] / cdfs[i,-1]

    colors = hex_cmap(cmap, n_frames+2)
    for i in range(n_frames):
        axes.plot(bin_edges[1:], cdfs[i,:], color=colors[i],
            linewidth=linewidth,
            label="$\Delta t = $ %.4f sec" % (frame_interval * (i+1)))

    axes.set_xlabel("2D radial jump ($\mu$m)", fontsize=fontsize)
    axes.set_ylabel("CDF")
    axes.legend(frameon=False, loc="lower right", prop={'size': 8})
    if not plot_max_r is None:
        axes.set_xlim((axes.get_xlim()[0], plot_max_r))

# rad_disp_histogram_2d(tracks, n_frames=4, bin_size=0.001, 
#     max_jump=5.0, pixel_size_um=0.160, n_gaps=0, use_entire_track=False,
#     max_jumps_per_track=10)

def plotRadialDisps(radial_disps, bin_edges, frame_interval=0.00548, plot=True):
    """
    Plot a set of radial displacement histograms as a simple line plot.

    args
    ----
        radial_disps        :   2D ndarray of shape (n_intervals, n_bins),
                                a set of radial displacement histograms 
                                for potentiall several frame intervals binned
                                spatially
        bin_edges           :   1D ndarray of shape (n_bins+1), the bin edge
                                definitions 
        frame_interval      :   float, seconds between frames
        plot                :   bool, immediately make a temporary plot to 
                                show to the user. Otherwise return the
                                Figure and Axes objects used to generate 
                                the image.

    returns
    -------
        if plot: 
            None

        else:
            (
                matplotlib.figure.Figure,
                1D array of matplotlib.axes.Axes
            )

    """
    if len(radial_disps.shape) == 1:
        radial_disps = np.asarray([radial_disps])
    assert radial_disps.shape[1] == (bin_edges.shape[0]-1)

    # Get the shape of the histograms
    n_intervals, n_bins = radial_disps.shape 

    # Get the centers of each bin
    bin_size = bin_edges[1] - bin_edges[0]
    bin_c = bin_edges[:-1] + bin_size/2.0

    # Set up the plot
    fig, ax = plt.subplots(n_intervals, 1, figsize=(4, 1.5*n_intervals),
        sharex=True)
    if n_intervals == 1:
        ax = np.asarray([ax])

    colors = hex_cmap("viridis", n_intervals*2)

    for interval in range(n_intervals):

        # Normalize
        pmf = radial_disps[interval,:] / radial_disps[interval,:].sum()

        # Plot
        ax[interval].plot(bin_c, pmf*bin_size, color=colors[interval],
            linestyle='-', linewidth=1.5,
            label="%.1f ms" % (frame_interval*1000*(interval+1)))
        ax[interval].legend(frameon=False, loc='upper right')

    # Axis labels
    ax[-1].set_xlabel("Radial displacement ($\mu$m)")
    for j in range(len(ax)):
        ax[j].set_ylabel("PMF")

    if plot:
        plt.show(); plt.close()
    else:
        return fig, ax 

def plotRadialDispsBar(radial_disps, bin_edges, frame_interval=0.00548,
    model=None, plot=True):
    """
    Plot a set of radial displacements as a bar graph. Also
    overlay a model as a line plot if desired.

    args
    ----
        radial_disps            :   2D ndarray of shape (n_intervals,
                                    n_bins), the radial displacements
        bin_edges               :   1D ndarray of shape (n_bins+1), bin
                                    edge definitions
        frame_interval          :   float, in sec
        model                   :   2D ndarray of shape (n_intervals,
                                    n_bins), model at each point
        plot                    :   bool, show immediately

    returns
    -------
        if plot:
            None

        else:
            (
                matplotlib.figure.Figure,
                array of matlotlib.axes.Axes
            )

    """
    # Check user inputs
    if len(radial_disps.shape) == 1:
        radial_disps = np.asarray([radial_disps])
    assert radial_disps.shape[1] == (bin_edges.shape[0]-1)

    # Check model, if using
    if not model is None:
        assert model.shape == radial_disps.shape

    # Get the shape of the histograms
    n_intervals, n_bins = radial_disps.shape 

    # Get the centers of each bin
    bin_size = bin_edges[1] - bin_edges[0]
    bin_c = bin_edges[:-1] + 0.5*bin_size 

    # Set up the plot
    fig, ax = plt.subplots(n_intervals, 1, figsize=(4, 1.5*n_intervals),
        sharex=True)
    if n_intervals == 1:
        ax = np.asarray([ax])

    colors = hex_cmap("viridis", n_intervals*2)

    for interval in range(n_intervals):

        # Normalize
        pmf = bin_size * radial_disps[interval,:] / \
            radial_disps[interval,:].sum()

        # Plot
        ax[interval].bar(bin_c, pmf, color=colors[interval],
            linestyle='-', linewidth=1, edgecolor='k',
            width=bin_size*0.8,
            label="%.1f ms" % (frame_interval*1000*(interval+1)))

        # Model definition, if using
        if not model is None:
            ax[interval].plot(bin_c, model[interval,:], color='k',
                linestyle='--', linewidth=1.5, label="Model")

        # Legend: show the frame interval 
        ax[interval].legend(frameon=False, loc='upper right')

    # Axis labels
    ax[-1].set_xlabel("Radial displacement ($\mu$m)")
    for j in range(len(ax)):
        ax[j].set_ylabel("PMF")

    if plot:
        plt.show(); plt.close()
    else:
        return fig, ax

###################
## MISCELLANEOUS ##
###################

def plot_pixel_mean_variance(means, variances, origin_files=None,
    model_gain=None, model_bg=None):
    """
    Plot pixel mean vs. variance for one or several movies,
    overlaying a linear gain model on top if desired.

    Best called through quot.read.measure_camera_gain.

    args
    ----
        means           :   list of 1D ndarray, the pixel 
                            means for each movie to plot
        variances       :   list of 1D ndarray, the pixel
                            variances for each movie to plot
        origin_files    :   list of str, the labels for each
                            element in *means* and *variances*
        model_gain      :   float, the camera gain 
        model_bg        :   float, the camera BG

    """
    # Check user inputs
    assert len(means) == len(variances)
    if not origin_files is None:
        assert len(origin_files) == len(means)

    # Plot setup
    fig, ax = plt.subplots(figsize=(3, 3))
    cmap = cm.get_cmap("gray", len(means)+2)
    colors = [mpl_colors.rgb2hex(cmap(i)) for i in range(len(means)+2)]

    # Plot each set of means and variances in a different color
    for i, (pixel_means, pixel_vars) in enumerate(zip(means, variances)):
        if not origin_files is None:
            label = origin_files[i]
        else:
            label = None 
        ax.scatter(pixel_means, pixel_vars, s=20, color=colors[i+1], label=label)

    # Plot the model overlay
    if (not model_gain is None) and (not model_bg is None):
        mean_min = min([j.min() for j in means])
        mean_max = max([j.max() for j in means])
        model_means = np.linspace(mean_min, mean_max, 101)
        model_vars = (model_means - model_bg) * model_gain 
        ax.plot(model_means, model_vars, linestyle='--', color='k', label="Model")

    # Limit the y extent if there are variance outliers (which is common)
    var_upper = np.percentile(np.concatenate(variances), 99.9)
    ax.set_ylim((0, var_upper*1.5))

    # Labels
    ax.legend(frameon=False, prop={'size': 6}, loc="upper left")
    ax.set_xlabel("Pixel mean (AU)")
    ax.set_ylabel("Pixel variance (AU$^{2}$)")

    plt.tight_layout(); plt.show(); plt.close()

##########################
## ANGULAR DISTRIBUTION ##
##########################

def angular_dist(axes, tracks, min_disp=0.2, delta=1, n_bins=50, norm=True,
    pixel_size_um=0.16, bottom=0.02, angle_ticks=True):
    """
    Plot the angular distribution of displacements for a 
    set of trajectories.

    Note that *axes* must be generated with projection = "polar".
    For instance, 

        fig = plt.figure()
        axes = fig.add_subplot(projection="polar")

    args
    ----
        axes        :   matplotlib.axes.Axes
        tracks      :   pandas.DataFrame
        min_disp    :   float, the minimum displacement required
                        to consider a displacement for an angle
                        in um
        delta       :   int, the number of subsequent jumps over
                        which to calculate the angle. For example,
                        if *delta* is 1, then angles are calculated
                        between subsequent jumps in a trajectory.
                        If *delta* is 2, then angles are calculated
                        between jump n and jump n+2, etc.
        n_bins      :   int, number of histogram bins
        norm        :   bool, normalize histogram
        pixel_size_um   :   float, size of pixels in um
        angle_ticks :   bool, do tick labels for the angles

    returns
    -------
        None, plots directly to *axes*

    """
    # Calculate angles
    angles = bond_angles(tracks, min_disp=min_disp/pixel_size_um, delta=delta)

    # Make a histogram
    bin_edges = np.linspace(0, np.pi, n_bins+1)
    bin_size = bin_edges[1] - bin_edges[0]
    bin_c = bin_edges[:-1] + bin_size * 0.5
    H = np.histogram(angles, bins=bin_edges)[0]

    if norm and H.sum() > 0:
        H = H.astype(np.float64) / H.sum()

    # Plot
    bars = axes.bar(bin_c, H, color="w", edgecolor="k",
        width=bin_size, bottom=bottom)
    axes.plot(bin_c, [bottom for i in range(n_bins)],
        color="k", linewidth=1) 
    yt = np.asarray(axes.get_yticks())
    axes.set_yticks(yt[yt>=bottom])
    axes.set_yticklabels([])
    axes.set_thetamin(0)
    axes.set_thetamax(180)
    axes.set_xticks([0, np.pi/3, 2*np.pi/3, np.pi])
    if not angle_ticks:
        axes.set_xticklabels([])
    axes.set_ylim((0, 1.1*axes.get_ylim()[1]))



###################################
## MULTI-FILE PLOTTING FUNCTIONS ##
###################################

def imlocdensities(*csv_files, out_png=None, filenames=False, **kwargs):
    """
    For each of a set of CSV files, make a KDE for localization
    density. This is a wrapper around imlocdensity().

    If *out_png* is passed, this function saves the result to a file.
    Otherwise it plots to the screen.

    args
    ----
        csv_files       :   variadic str, a set of CSV files
        out_png         :   str, save filename
        filenames       :   str, include the filenames as the
                            plot title
        kwargs          :   keyword arguments to imlocdensity().
                            See that function's docstring for
                            more info. These include: ymax, xmax,
                            bin_size, kernel, vmax_perc, cmap,
                            pixel_size_um, scalebar

    """
    n = len(csv_files)
    if n == 0:
        print("quot.plot.imlocdensities: no files passed")
        return 

    # Number of subplots on an edge
    m = int(np.ceil(np.sqrt(n)))

    mx = m 
    my = n // m + 1

    # Lay out the main plot
    fig, ax = plt.subplots(my, mx, figsize=(3*mx, 3*my))

    # Make localization density plots for each file
    for i, csv_file in enumerate(csv_files):
        tracks = pd.read_csv(csv_file)
        imlocdensity(ax[i//mx, i%mx], tracks, **kwargs)

        # Set filename as plot title if desired
        if filenames:
            ax[i//mx, i%mx].set_title(csv_file, fontsize=8)

    # Set the remainder of the plots to invisible
    for i in range(n, my*mx):
        kill_ticks(ax[i//mx, i%mx])

    # Save or show
    if not out_png is None:
        wrapup(out_png)
    else:
        plt.tight_layout(); plt.show(); plt.close()

def locs_per_frame_files(*csv_files, out_png=None, **kwargs):
    """
    Given a set of trajectory CSVs, make a plot where each subpanel
    shows the number of localizations in that file as a function of
    time. 

    args
    ----
        csv_files       :   variadic str, paths to CSVs
        out_png         :   str, save file
        kwargs          :   to locs_per_frame()

    """
    n = len(csv_files)
    if n == 0:
        print("quot.plot.locs_per_frame_files: no files passed")
        return 

    # Number of subplots on an edge
    m = int(np.ceil(np.sqrt(n)))

    mx = m 
    my = n//m + 1

    # Make the plot
    fig, ax = plt.subplots(my, mx, figsize=(mx*3, my*1.5))
    for i, csv_file in enumerate(csv_files):
        tracks = pd.read_csv(csv_file)
        locs_per_frame(ax[i//mx, i%mx], tracks, title=csv_file, **kwargs)

    # Set unused subplots to blank
    for i in range(n, my*mx):
        kill_ticks(ax[i//mx, i%mx])

    # Save or show
    if not out_png is None:
        wrapup(out_png)
    else:
        plt.tight_layout(); plt.show(); plt.close()

def jump_cdf_files(*csv_files, out_png=None, **kwargs):
    """
    For each of a set of trajectory CSVs, plot the jump
    length CDFs.

    args
    ----
        csv_files       :   list of str, paths to CSV files
        out_png         :   str, output plot file
        kwargs          :   to plot_jump_cdfs(). These include:
                            n_frames, pixel_size_um, frame_interval,
                            fontsize, cmap, start_frame,
                            max_jumps_per_track, use_entire_track,
                            linewidth

    returns
    -------
        None; either plots to screen or saves depending
            on whether *out_png* is set

    """
    n = len(csv_files)
    if n == 0:
        print("quot.plot.jump_cdfs_files: no files passed")
        return 

    m = int(np.ceil(np.sqrt(n)))
    mx = m 
    my = n//m + 1

    fig, ax = plt.subplots(my, mx, figsize=(4*mx, 2*my))
    for i, csv_file in enumerate(csv_files):
        tracks = pd.read_csv(csv_file)
        plot_jump_cdfs(ax[i//mx, i%mx], tracks, **kwargs)
        ax[i//mx, i%mx].set_title(csv_file, fontsize=10)

    for i in range(n, my*mx):
        kill_ticks(ax[i//mx, i%mx])

    if not out_png is None:
        wrapup(out_png)
    else:
        plt.tight_layout(); plt.show(); plt.close()

def angular_dist_files(*csv_files, out_png=None, start_frame=0,
    filenames=False, **kwargs):
    """
    For each of a set of trajectory CSVs, plot the angular
    distribution of displacements.

    args
    ----
        csv_files       :   variadic str, paths to CSV files
        start_frame     :   int, the first frame in the file
                            consider
        out_png         :   str, save path
        kwargs          :   to angular_dist(). These include:
                            min_disp (um), n_bins, norm,
                            pixel_size_um

    returns
    -------
        None

    """
    n = len(csv_files)
    if n == 0:
        print("quot.plot.angular_dist_files: no files passed")
        return 

    m = int(np.ceil(np.sqrt(n)))
    mx = m 
    my = n//m + 1

    fig = plt.figure(figsize=(3*mx, 2*my))
    axes = []
    for i in range(n):
        ax = fig.add_subplot(my, mx, i+1,
            projection="polar")
        axes.append(ax)

    for i, csv_file in enumerate(csv_files):
        tracks = pd.read_csv(csv_file)
        tracks = tracks[tracks["frame"] >= start_frame]
        angular_dist(axes[i], tracks, **kwargs)
        if filenames:
            axes[i].set_title(csv_file, fontsize=10)

    if not out_png is None:
        wrapup(out_png)
    else:
        plt.tight_layout(); plt.show(); plt.close()


#######################
## RELATED UTILITIES ##
#######################

def track_length(tracks):
    """
    Given a set of trajectories in DataFrame format, create
    a new columns ("track_length") with the length of the 
    corresponding trajectory in frames.

    args
    ----
        tracks      :   pandas.DataFrame

    returns
    -------
        pandas.DataFrame, with the new column

    """
    if "track_length" in tracks.columns:
        tracks = tracks.drop("track_length", axis=1)
    tracks = tracks.join(
        tracks.groupby("trajectory").size().rename("track_length"),
        on="trajectory"
    )
    return tracks 

def rad_disp_histogram_2d(tracks, n_frames=4, bin_size=0.001, 
    max_jump=5.0, pixel_size_um=0.160, n_gaps=0, use_entire_track=False,
    max_jumps_per_track=10):
    """
    Compile a histogram of radial displacements in the XY plane for 
    a set of trajectories ("tracks").

    Identical with strobemodels.utils.rad_disp_histogram_2d.

    args
    ----
        tracks          :   pandas.DataFrame
        n_frames        :   int, the number of frame delays to consider.
                            A separate histogram is compiled for each
                            frame delay.
        bin_size        :   float, the size of the bins in um. For typical
                            experiments, this should not be changed because
                            some diffusion models (e.g. Levy flights) are 
                            contingent on the default binning parameters.
        max_jump        :   float, the max radial displacement to consider in 
                            um
        pixel_size_um   :   float, the size of individual pixels in um
        n_gaps          :   int, the number of gaps allowed during tracking
        use_entire_track:   bool, use every displacement in the dataset
        max_jumps_per_track:   int, the maximum number of displacements
                            to consider per trajectory. Ignored if 
                            *use_entire_track* is *True*.

    returns
    -------
        (
            2D ndarray of shape (n_frames, n_bins), the distribution of 
                displacements at each time point;
            1D ndarray of shape (n_bins+1), the edges of each bin in um
        )

    """
    # Sort by trajectory, then frame
    tracks = tracks.sort_values(by=["trajectory", "frame"])

    # Assign track lengths
    if "track_length" not in tracks.columns:
        tracks = track_length(tracks)

    # Filter out unassigned localizations and singlets
    T = tracks[
        np.logical_and(tracks["trajectory"]>=0, tracks["track_length"]>1)
    ][["frame", "trajectory", "y", "x"]]

    # Convert to ndarray for speed
    T = np.asarray(T[["frame", "trajectory", "y", "x", "trajectory"]]).astype(np.float64)

    # Sort first by track, then by frame
    T = T[np.lexsort((T[:,0], T[:,1])), :]

    # Convert from pixels to um
    T[:,2:4] = T[:,2:4] * pixel_size_um 

    # Format output histogram
    bin_edges = np.arange(0.0, max_jump+bin_size, bin_size)
    n_bins = bin_edges.shape[0]-1
    H = np.zeros((n_frames, n_bins), dtype=np.int64)

    # Consider gap frames
    if n_gaps > 0:

        # The maximum trajectory length, including gap frames
        max_len = (n_gaps + 1) * n_frames + 1

        # Consider every shift up to the maximum trajectory length
        for l in range(1, max_len+1):

            # Compute the displacement for all possible jumps
            diff = T[l:,:] - T[:-l,:]

            # Map the trajectory index corresponding to the first point in 
            # each trajectory
            diff[:,4] = T[l:,1]

            # Only consider vectors between points originating from the same track
            diff = diff[diff[:,1] == 0.0, :]

            # Look for jumps corresponding to each frame interval being considered
            for t in range(1, n_frames+1):

                # Find vectors that match the delay being considered
                subdiff = diff[diff[:,0] == t, :]

                # Only consider a finite number of displacements from each trajectory
                if not use_entire_track:
                    _df = pd.DataFrame(subdiff[:,4], columns=["traj"])
                    _df["ones"] = 1
                    _df["index_in_track"] = _df.groupby("traj")["ones"].cumsum() 
                    subdiff = subdiff[np.asarray(_df["index_in_track"]) <= max_jumps_per_track, :]

                # Calculate radial displacements
                r_disps = np.sqrt((subdiff[:,2:4]**2).sum(axis=1))
                H[t-1,:] = H[t-1,:] + np.histogram(r_disps, bins=bin_edges)[0]

    # No gap frames
    else:

        # For each frame interval and each track, calculate the vector change in position
        for t in range(1, n_frames+1):
            diff = T[t:,:] - T[:-t,:]

            # Map trajectory indices back to the first localization of each trajectory
            diff[:,4] = T[t:,1]

            # Only consider vectors between points originating in the same track
            diff = diff[diff[:,1] == 0.0, :]

            # Only consider vectors that match the delay being considered
            diff = diff[diff[:,0] == t, :]

            # Only consider a finite number of displacements from each trajectory
            if not use_entire_track:
                _df = pd.DataFrame(diff[:,4], columns=["traj"])
                _df["ones"] = 1
                _df["index_in_track"] = _df.groupby("traj")["ones"].cumsum()
                diff = diff[np.asarray(_df["index_in_track"]) <= max_jumps_per_track, :]

            # Calculate radial displacements
            r_disps = np.sqrt((diff[:,2:4]**2).sum(axis=1))
            H[t-1,:] = np.histogram(r_disps, bins=bin_edges)[0]

    return H, bin_edges 

def coarsen_histogram(jump_length_histo, bin_edges, factor):
    """
    Given a jump length histogram with many small bins, aggregate into a 
    histogram with a small number of larger bins.

    This is useful for visualization.

    args
    ----
        jump_length_histo       :   2D ndarray, the jump length histograms
                                    indexed by (frame interval, jump length bin)
        bin_edges               :   1D ndarray, the edges of each jump length
                                    bin in *jump_length_histo*
        factor                  :   int, the number of bins in the old histogram
                                    to aggregate for each bin of the new histogram

    returns
    -------
        (
            2D ndarray, the aggregated histogram,
            1D ndarray, the edges of each jump length bin the aggregated histogram
        )

    """
    # Get the new set of bin edges
    n_frames, n_bins_orig = jump_length_histo.shape 
    bin_edges_new = bin_edges[::factor]
    n_bins_new = bin_edges_new.shape[0] - 1

    # May need to truncate the histogram at the very end, if *factor* doesn't
    # go cleanly into the number of bins in the original histogram
    H_old = jump_length_histo[:, (bin_edges<bin_edges_new[-1])[:-1]]

    # Aggregate the histogram
    H = np.zeros((n_frames, n_bins_new), dtype=jump_length_histo.dtype)
    for j in range(factor):
        H = H + H_old[:, j::factor]

    return H, bin_edges_new 

def bond_angles(tracks, min_disp=0.2, delta=1):
    """
    Return the set of angles between jumps originating from 
    trajectories. Angles between pi and 2 * pi are reflected onto
    the interval 0, pi.

    args
    ----
        tracks      :   pandas.DataFrame
        min_disp    :   float, pixels. Discard displacements less than
                        this displacement. This prevents us from
                        being biased by localization error.
        delta       :   int, the proximity of the two jumps with
                        respect to which the angle is calculated.
                        For example, if *delta* is 1, then the angle
                        between subsequent jumps is calculated. 
                        If *delta* is 2, then the angle between 
                        jump n and jump n+2 is calculated, and so on.

    returns
    -------
        1D ndarray of shape (n_angles,), the observed
            angles in radians (from 0 to pi)

    """
    assert delta > 0, "delta must be greater than 0 (passed: {})".format(delta)

    # Get the set of all trajectories with at least three points
    # as an ndarray
    tracks = track_length(tracks)
    T = np.asarray(
        tracks[(tracks["trajectory"] >= 0) & (tracks["track_length"] > (1+delta))][
            ["trajectory", "frame", "y", "x"]
        ]
    )
    if T.shape[0] == 0:
        return np.nan

    # Ascending trajectory and frame indices
    tracks = tracks.sort_values(by=["trajectory", "frame"])

    # Unique trajectory indices
    traj_indices = np.unique(T[:, 0])

    # Largest number of possible angles
    n_angles = T.shape[0] - 2 * len(traj_indices)

    # Output array
    angles = np.zeros(n_angles, dtype="float64")

    # Iterate through all trajectories
    c = 0
    for i, j in enumerate(traj_indices):

        # The set of y, x points corresponding to this trajectory
        traj = T[T[:, 0] == j, 2:]

        # Jumps between subsequent frames
        disps = traj[1:, :] - traj[:-1, :]

        # Jump moduli
        mags = np.sqrt((disps ** 2).sum(axis=1))

        # Angles between subsequent jumps
        traj_angles = (disps[delta:, :] * disps[:-delta, :]).sum(axis=1) / (mags[delta:] * mags[:-delta])

        # Only take angles above a given displacement, if desired
        traj_angles = traj_angles[(mags[delta:] >= min_disp) & (mags[:-delta] >= min_disp)]

        # Aggregate
        n_traj_angles = traj_angles.shape[0]
        angles[c : c + n_traj_angles] = traj_angles
        c += n_traj_angles

    # We'll lose some angles because of the min_disp cutoff
    angles = angles[:c]

    # Some floating point errors occur here - values slightly
    # greater than 1.0 or less than -1.0
    angles[angles > 1.0] = 1.0 
    angles[angles < -1.0] = -1.0 

    return np.arccos(angles[~pd.isnull(angles)])



__version__ = "0.0.1"


#from ._widget import PalmariWidget
from .data_structure.experiment import Experiment
from .data_structure.acquisition import Acquisition
from .processing.image_pipeline import ImagePipeline

# from .processing.image_pipeline_widget import ImagePipelineWidget



from .base import Detector, SubpixelLocalizer
from ...quot.core import localize_frame, detect
import pandas as pd
import numpy as np
from typing import Dict
import enum

detection_methods_dict = {
    "log": "Laplacian of Gaussians",
    "llr": "Likelihood ratio",
}


class BaseDetector(Detector):
    def __init__(
        self, k: float = 1.5, w: int = 21, t: float = 1.0, method: str = "llr"
    ):
        self.k = k
        self.w = w
        self.t = t
        self.method = method

    def detect_frame(self, img: np.array) -> pd.DataFrame:
        return detect(img, t=self.t, k=self.k, method=self.method, w=self.w)

    @property
    def name(self):
        return "Log-likelihood detector"

    widget_types = {"t": "FloatSlider", "method": "ComboBox"}
    widget_options = {
        "t": {
            "step": 0.1,
            "min": 0.5,
            "max": 3.0,
            "label": "Threshold",
            "readout": True,
        },
        "k": {
            "step": 0.1,
            "label": "Spot size (px)",
            "min": 0.0,
        },
        "w": {
            "step": 2,
            "label": "Window size (px)",
            "min": 5,
            "tooltip": "Size of the square window used for testing.",
        },
        "method": {
            "label": "Method",
            "choices": [
                ("Log-likelihood ratio", "llr"),
                ("Laplacian of Gaussians", "log"),
            ],
        },
    }


class RadialLocalizer(SubpixelLocalizer):

    cols_dtype = {
        "y": float,
        "x": float,
        "I0": float,
        "bg": float,
        "error_flag": bool,
        "snr": float,
    }

    widget_options = {
        "window_size": {"label": "Window size (px)", "step": 1},
    }

    def __init__(
        self,
        window_size: int = 9,
    ):
        self.window_size = window_size

    @property
    def name(self):
        return "Radial symmetry localizer"

    def localize_frame(
        self, img: np.array, detections: np.array
    ) -> pd.DataFrame:
        return localize_frame(
            img.T,
            detections,
            method="radial_symmetry",
            sigma=1.0,
            window_size=self.window_size,
        )


class MaxLikelihoodLocalizer(SubpixelLocalizer):

    cols_dtype = {
        "x": float,
        "y": float,
        "frame": int,
        "I0": float,
        "bg": float,
        "y_err": float,
        "x_err": float,
        "I0_err": float,
        "bg_err": float,
        "H_det": float,
        "error_flag": int,
        "snr": float,
        "rmse": float,
    }

    widget_types = {"method": "ComboBox"}
    widget_options = {
        "method": {
            "label": "Noise",
            "value": "ls_int_gaussian",
            "choices": [
                ("Gaussian", "ls_int_gaussian"),
                ("Poisson", "poisson_int_gaussian"),
            ],
            "tooltip": "Assumed type of background noise",
        },
        "window_size": {"label": "Window size (px)", "step": 1},
        "sigma": {"label": "Spot size (px)", "step": 0.1},
    }

    def __init__(
        self,
        method: str = "ls_int_gaussian",
        window_size: int = 9,
        sigma: float = 1.5,
    ):
        self._loc_params = {
            "ridge": 0.0001,
            "convergence": 0.0001,
            "divergence": 1.0,
            "max_iter": 10,
            "damp": 0.3,
            "camera_bg": 0.0,
            "camera_gain": 1.0,
        }
        self.method = method
        self.window_size = window_size
        self.sigma = sigma

    @property
    def name(self):
        return "Max-likelihood subpixel localizer"

    def localize_frame(
        self, img: np.array, detections: np.array
    ) -> pd.DataFrame:
        return localize_frame(
            img.T,
            detections,
            method=self.method,
            sigma=self.sigma,
            window_size=self.window_size,
            **self._loc_params
        )



import pandas as pd
import numpy as np
from .base import Tracker
from typing import Dict
import logging


class DiffusionTracker(Tracker):
    def __init__(
        self,
        max_diffusivity: float = 5.0,
        max_blinks: int = 0,
        # For the diffusion weight matrix
        d_bound_naive: float = 0.1,  # naive estimate of diffusion
        init_cost: float = 50.0,  # Cost of starting a new trajectory when reconnections are available in the search radius
        y_diff: float = 0.9,  # relative importance of the trajectory's history in diffusion estimation
    ):
        # Attributes will automatically be detected as parameters of the step and stored/loaded.
        # Parameters must have default values
        self.max_diffusivity = max_diffusivity
        self.y_diff = y_diff
        self.init_cost = init_cost
        self.d_bound_naive = d_bound_naive
        self.max_blinks = max_blinks

    def track(self, locs: pd.DataFrame):
        # This is where the actual tracking happen.
        from ...quot.core import track

        delta_t = self.estimate_delta_t(locs)  # This is a Tracker's method.
        dim = 2
        max_radius = np.sqrt(2 * dim * self.max_diffusivity * delta_t)
        logging.info("Max radius is %.2f" % max_radius)
        locs["n"] = track(
            locs,
            method="diffusion",
            search_radius=max_radius,
            pixel_size_um=1.0,
            frame_interval=delta_t,
            max_blinks=self.max_blinks,
            y_diff=self.y_diff,
            d_bound_naive=self.d_bound_naive,
            init_cost=self.init_cost,
        )["trajectory"]
        return locs

    @property
    def name(self):
        # This is for printing
        return "Diffusion tracker"

    # The following dicts are used when setting the parameters through a graphic interface, using open_in_napari()
    widget_types = {
        "max_diffusivity": "FloatSpinBox",
        "max_blinks": "SpinBox",
        "d_bound_naive": "FloatSpinBox",
        "init_cost": "FloatSpinBox",
        "y_diff": "FloatSlider",
    }
    # For details about widget types, see https://napari.org/magicgui/
    widget_options = {
        "max_diffusivity": {
            "step": 1.0,
            "tooltip": "Assumed maximum diffusivity (in microns per square second).\nThis is used in conjunction with the Time delta to set the maximal distance between consecutive localizations",
            "label": "D_max (um^2/s)",
            "min": 0.0,
        },
        "max_blinks": {
            "step": 1,
            "tooltip": "Maximum number of tolerated blinks (i.e. number of frames during which a particle can 'disappear').",
            "label": "Max blinks",
            "min": 0,
            "max": 2,
        },
        "y_diff": {
            "step": 0.01,
            "min": 0.0,
            "max": 1.0,
            "label": "Past vs naive",
            "tooltip": "Relative weight of a trajectory's past in the estimation of its diffusivity",
        },
        "d_bound_naive": {
            "step": 0.1,
            "tooltip": "First guess of a trajectory's diffusivity",
            "label": "D_0 (um^2/s)",
            "min": 0.0,
        },
        "init_cost": {
            "step": 10,
            "label": "Initialization cost",
            "min": 1,
            "tooltip": "Cost of initializing a new trajectory when a reconnection is available in the search radius",
        },
    }


class EuclideanTracker(Tracker):
    def __init__(
        self,
        max_diffusivity: float = 5.0,
        max_blinks: int = 0,
        scale: float = 1.0,  # scaling factor between distances and weights
        init_cost: float = 50.0,  # Cost of starting a new trajectory when reconnections are available in the search radius
    ):
        # Attributes will automatically be detected as parameters of the step and stored/loaded.
        # Parameters must have default values
        self.max_diffusivity = max_diffusivity
        self.init_cost = init_cost
        self.max_blinks = max_blinks
        self.scale = scale

    def track(self, locs: pd.DataFrame):
        # This is where the actual tracking happen.
        from ...quot.core import track

        delta_t = self.estimate_delta_t(locs)  # This is a Tracker's method.
        dim = 2
        max_radius = np.sqrt(2 * dim * self.max_diffusivity * delta_t)
        logging.info("Max radius is %.2f" % max_radius)
        locs["n"] = track(
            locs,
            method="euclidean",
            search_radius=max_radius,
            pixel_size_um=1.0,
            frame_interval=delta_t,
            max_blinks=self.max_blinks,
            scale=self.scale,
            init_cost=self.init_cost,
        )["trajectory"]
        return locs

    @property
    def name(self):
        # This is for printing
        return "Euclidean tracker"

    # The following dicts are used when setting the parameters through a graphic interface, using open_in_napari()
    widget_types = {
        "max_diffusivity": "FloatSpinBox",
        "scale": "FloatSpinBox",
        "init_cost": "FloatSpinBox",
        "max_blinks": "SpinBox",
    }
    # For details about widget types, see https://napari.org/magicgui/
    widget_options = {
        "max_diffusivity": {
            "step": 1.0,
            "tooltip": "Assumed maximum diffusivity (in microns per square second).\nThis is used in conjunction with the Time delta to set the maximal distance between consecutive localizations",
            "label": "D_max (um^2/s)",
            "min": 0.0,
        },
        "scale": {
            "step": 0.5,
            "label": "Weight scale",
            "min": 0.1,
            "tooltip": "Scaling factor between distance (in um) and coefficient of the corresponding assignment in the weight matrix",
        },
        "init_cost": {
            "step": 10,
            "label": "Initialization cost",
            "min": 1,
            "tooltip": "Cost of initializing a new trajectory when a reconnection is available in the search radius",
        },
        "max_blinks": {
            "step": 1,
            "tooltip": "Maximum number of tolerated blinks (i.e. number of frames during which a particle can 'disappear').",
            "label": "Max blinks",
            "min": 0,
            "max": 2,
        },
    }



#!/usr/bin/env python
"""
quot/read.py -- image file readers

"""
# ndarrays
import numpy as np 

# File paths
import os

# Read config files in TOML format
import toml 

# Underlying file readers for each file format

def read_config(path):
    """
    Read a config file.

    args
        path    :   str

    returns
        dict

    """
    assert os.path.isfile(path), \
        "read_config: path %s does not exist" % path
    return toml.load(path)

def save_config(path, config):
    """
    Save data in a dict to a TOML file.

    args
        path    :   str
        config  :   dict

    """
    with open(path, 'w') as o:
        toml.dump(config, o)

class ImageReader(object):
    """
    An image file reader that aims to provide a single API for
    several image types, relying on other packages to read 
    individual file types. Focuses especially on methods to 
    return temporal frames.

    *start* and *stop* are the frame limits when iterating over
    the ImageReader. They do not prohibit the user from accessing
    frames outside this range with ImageReader.get_frame() or 
    related methods.

    init
        path        :   str
        start       :   int, limits of iteration. If not set
                        defaults to 0.
        stop        :   int, limits of iteration. If not set
                        defaults to the last frame.
        subregion   :   keys y0, y1, x0, x1 (all int), the 
                        subregion to use for iteration if 
                        desired

    """
    def __init__(self, path, start=None, stop=None, **subregion):

        assert os.path.isfile(path), "ImageReader.__init__: " \
            "path %s does not exist" % path 
        self.path = path

        if '.nd2' in path:
            self.ext = '.nd2'
            from nd2reader import ND2Reader 
            self._reader = ND2Reader(path)
        elif ('.tif' in path) or ('.tiff' in path):
            self.ext = '.tif'
            import tifffile
            self._reader = tifffile.TiffFile(path)
        #elif ('.czi' in path):
        #    self.ext = '.czi'
        #    self._reader = CziFile(path)
        else:
            raise RuntimeError("Image format %s not recognized" % \
                os.path.splitext(path)[1])

        # Record movie shape
        self.n_frames, self.height, self.width = self.shape 

        # Set defaults for iteration 
        if start is None:
            start = 0
        if stop is None:
            stop = self.n_frames 
        self.start = start 
        self.stop = stop 

    def __iter__(self):
        self._c = self.start 
        return self 

    def __next__(self):
        if self._c < self.stop:
            self._c += 1
            return self.get_frame(self._c-1)
        else:
            raise StopIteration

    def __enter__(self):
        return self 

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """ Close the filestream. """
        self._reader.close()

    def _frame_valid(self, frame_index):
        """
        True of the corresponding frame exists.

        """
        return (frame_index>=0) and (frame_index<self.n_frames)

    def _frame_range_valid(self, start, stop):
        """
        True if all frames in the range exist.

        """
        return start<=self.n_frames and \
            stop<=self.n_frames and \
            (start<stop)

    def _subregion_valid(self, **subregion):
        """
        True if the subregion lies within the limits of 
        the movie.

        """
        return subregion.get('y0',0)>=0 and \
            subregion.get('x0',0)>=0 and \
            subregion.get('y1',self.height)<=self.height and \
            subregion.get('x1',self.width)<=self.width

    def _process_frame_range(self, **kwargs):
        """
        Truncate a frame range into the permissible limits 
        for this movie.

        kwargs
            start, stop     :   int, the start and stop frame
                                indices for the range. Added
                                if they don't exist.

        returns
            dict, modified kwargs

        """
        kwargs['start'] = max(0, kwargs.get('start', 0))
        kwargs['stop'] = min(self.n_frames, kwargs.get('stop', self.n_frames))
        return kwargs 

    def _process_subregion(self, **kwargs):
        """
        Truncate a subregion into the permissible limits for this
        movie.

        kwargs
            y0, y1, x0, x1  :   int, subregion limits

        returns
            dict, modified subregion kwargs

        """
        kwargs['y0'] = max(kwargs.get('y0', 0), 0)
        kwargs['x0'] = max(kwargs.get('x0', 0), 0)
        kwargs['y1'] = min(kwargs.get('y1', self.height), self.height)
        kwargs['x1'] = min(kwargs.get('x1', self.width), self.width)
        return kwargs

    def _process_subregion_range(self, **kwargs):
        """
        Similar to _process_subregion(), but also specifies the 
        *start* and *stop* kwargs for multiple frames.

        args
            y0, y1, x0, x1, start, stop :   int, subregion limits

        returns
            dict of keys y0, y1, x0, x1, start, stop

        """
        return self._process_frame_range(**self._process_subregion(**kwargs))

    @property 
    def dtype(self):
        """
        Return the dtype for the arrays produced by 
        self.get_frame() and related methods.

        returns
            str

        """
        if hasattr(self, "_dtype"):
            return self._dtype

        self._dtype = self.get_frame(0).dtype
        return self._dtype 

    @property 
    def shape(self):
        """
        Return the number of frames, height, and width of the 
        image movie.

        returns
            (int, int, int)

        """
        if hasattr(self, "_shape"):
            return self._shape 

        if self.ext == ".nd2":
            self._shape = (
                self._reader.metadata["total_images_per_channel"],
                self._reader.metadata["height"],
                self._reader.metadata["width"],
            )
        elif self.ext == ".tif":
            self._shape = (
                len(self._reader.pages),
                *self._reader.pages[0].shape, 
            )
        return self._shape 

    def get_frame(self, frame_index, c=0):
        """
        Return a single frame from the file.

        args
            frame_index     :   int
            c               :   int, channel index. Currently only 
                                implemented for ND2 files.

        returns
            2D ndarray (YX)

        """
        assert self._frame_valid(frame_index)
        if (self.ext == ".nd2") and (c == 0):
            return self._reader.get_frame_2D(t=frame_index)
        elif (self.ext == ".nd2") and (c > 0):
            return self._reader.get_frame_2D(t=frame_index, c=c)
        elif (self.ext == ".tif"):
            return self._reader.pages[frame_index].asarray()

    def get_frame_range(self, start, stop):
        """
        Get a contiguous range of frames as a 3D stack.

        args
            start, stop     :   int

        returns
            3D ndarray (TYX)

        """
        kwargs = self._process_frame_range(start=start, stop=stop)
        n = kwargs['stop'] - kwargs['start']
        result = np.empty((n, self.height, self.width), dtype=self.dtype)
        for i, frame_index in enumerate(range(kwargs['start'], kwargs['stop'])):
            result[i,:,:] = self.get_frame(frame_index)
        return result 

    def get_subregion(self, frame_index, **subregion):
        """
        Get a subregion of a single frame.

        args
            frame_index     :   int
            subregion       :   keys y0, y1, x0, x1,
                                the subregion limits

        returns
            3D ndarray (TYX)

        """
        subregion = self._process_subregion(**subregion)
        return self.get_frame(frame_index)[
            subregion['y0']:subregion['y1'],
            subregion['x0']:subregion['x1'],
        ]

    def get_subregion_range(self, **kwargs):
        """
        Get a subregion of multiple frames.

        kwargs
            y0, y1, x0, x1, start, stop :   int, the 
                subregion limits

        returns
            3D ndarray (TYX)

        """
        kwargs = self._process_subregion_range(**kwargs)
        T = kwargs['stop'] - kwargs['start']
        N = kwargs['y1'] - kwargs['y0']
        M = kwargs['x1'] - kwargs['x0']
        result = np.empty((T, N, M), dtype=self.dtype)
        for j, frame_index in enumerate(range(kwargs['start'], kwargs['stop'])):
            try:
                result[j,:,:] = self.get_frame(frame_index)[
                    kwargs['y0']:kwargs['y1'],
                    kwargs['x0']:kwargs['x1']
                ]
            except:
                result[j,:,:] = 0 
        return result 

    def max_int_proj(self, **kwargs):
        """
        Take a maximum intensity projection of the movie.

        kwargs
            start, stop :   int, the first and last frames to use

        returns
            2D ndarray (YX)

        """
        kwargs = self._process_frame_range(**kwargs)
        I = self.get_frame(kwargs['start'])
        for i in range(kwargs['start']+1, kwargs['stop']):
            I = np.maximum(I, self.get_frame(i))
        return I 

    def sum_proj(self, **kwargs):
        """
        Take a sum projection of the movie.

        kwargs
            start, stop :   int, the first and last frames to use

        returns
            2D ndarray (YX), dtype float64

        """
        kwargs = self._process_frame_range(**kwargs)
        I = self.get_frame(kwargs['start']).astype(np.float64)
        for i in range(kwargs['start']+1, kwargs['stop']):
            I += self.get_frame(i).astype(np.float64)
        return I 

    def min_max(self, **kwargs):
        """
        Get the minimum and maximum pixel intensities for a 
        frame range.

        kwargs
            start, stop :   int, the first and last frames to use

        returns
            (int min, int max)

        """
        kwargs = self._process_frame_range(**kwargs)
        cmax, cmin = 0, 1e10
        for i in range(kwargs['start'], kwargs['stop']):
            I = self.get_frame(i)
            cmax = max(cmax, I.max())
            cmin = min(cmin, I.min())
        return cmin, cmax 

    def imread(self):
        """
        Read the entire stack into memory.

        returns
            3D ndarray (TYX)

        """
        if self.ext == '.nd2':
            return self.get_frame_range(0, self.n_frames)
        elif self.ext == '.tif':
            return self._reader.asarray()

    def imsave(self, path, **kwargs):
        """
        Save part of the movie to a TIF.

        args
            path    :   str, out TIF
            kwargs  :   subregion kwargs

        """
        tifffile.imsave(path, 
            self.get_subregion_range(**kwargs).astype('uint16'))




from napari import Viewer
import os

from .processing import ImagePipeline
from .processing.steps import *
from qtpy.QtWidgets import (
    QDialog,
    QErrorMessage,
    QLineEdit,
    QPushButton,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
)
from qtpy.QtCore import Signal
from magicgui.widgets import FileEdit, FloatSpinBox, LineEdit, ProgressBar
from threading import Thread


class PipelineRunner(QWidget):

    progressChanged = Signal(int)
    progressRunning = Signal(bool)
    nFilesChanged = Signal(int)
    nameProcessingFile = Signal(str)

    def __init__(self):
        super().__init__()
        self.setup_ui()
        self.running = False
        self.pixel_size = 0.097
        self.DT = 0.03
        self.file_pattern = ".tif"

        def updatePgBar(s):
            self.pgbar.value = s

        def updateNFiles(s):
            self.pgbar.max = s

        def showPgBar(s):
            self.pgbar.visible = s
            self.file_label.setVisible(s)
            self.running = s
            self.enable_run_button()

        def updateFileName(s):
            self.file_label.setText(s)

        self.progressChanged.connect(updatePgBar)
        self.progressRunning.connect(showPgBar)
        self.nFilesChanged.connect(updateNFiles)
        self.nameProcessingFile.connect(updateFileName)

    def setup_ui(self):
        main_layout = QVBoxLayout()
        main_layout.addLayout(self.choose_pipeline_layout())
        main_layout.addLayout(self.choose_input_folder_layout())
        main_layout.addLayout(self.choose_export_folder_layout())

        main_layout.addLayout(self.exp_params_layout())

        self.run_button = QPushButton("Process (?) files !")
        self.run_button.setEnabled(False)
        self.run_button.clicked.connect(self.run_pipeline)
        main_layout.addWidget(self.run_button)

        self.pgbar = ProgressBar(min=0, visible=False)
        main_layout.addWidget(self.pgbar._widget._qwidget)
        self.file_label = QLabel("")
        main_layout.addWidget(self.file_label)

        self.setLayout(main_layout)

    ###
    # LAYOUT PARTS
    ###

    def choose_pipeline_layout(self):

        main_layout = QHBoxLayout()
        main_layout.addWidget(QLabel("Pipeline :"))
        self.select_file_widget = FileEdit(
            mode="r",
            filter="*.yaml",
            label="Pipeline file",
            name="Pipeline file",
        )
        self.select_file_widget.changed.connect(self.load_pipeline)
        main_layout.addWidget(self.select_file_widget._widget._qwidget)
        return main_layout

    def choose_input_folder_layout(self):

        main_layout = QHBoxLayout()
        main_layout.addWidget(QLabel("Process folder :"))
        self.select_input_folder_widget = FileEdit(
            mode="d",
            label="Process folder...",
            name="Process folder...",
        )
        self.select_input_folder_widget.changed.connect(self.setInputFolder)
        main_layout.addWidget(self.select_input_folder_widget._widget._qwidget)
        return main_layout

    def choose_export_folder_layout(self):

        main_layout = QHBoxLayout()
        main_layout.addWidget(QLabel("Save to folder:"))
        self.select_export_folder_widget = FileEdit(
            mode="d",
            label="Save to folder...",
            name="Save to folder...",
        )
        self.select_export_folder_widget.changed.connect(self.setExportFolder)
        main_layout.addWidget(
            self.select_export_folder_widget._widget._qwidget
        )
        return main_layout

    def exp_params_layout(self):

        DT_layout = QHBoxLayout()
        DT_layout.addWidget(QLabel("Exposure (in s.):"))
        dt_widget = FloatSpinBox(value=0.03, min=0.0, step=0.001)
        dt_widget.changed.connect(self.setExpDT)
        DT_layout.addWidget(dt_widget._widget._qwidget)

        pixel_layout = QHBoxLayout()
        pixel_layout.addWidget(QLabel("Pixel size (in um):"))
        pixel_widget = FloatSpinBox(value=0.097, min=0.0, step=0.001)
        pixel_widget.changed.connect(self.setExpPixel)
        pixel_layout.addWidget(pixel_widget._widget._qwidget)

        pattern_layout = QHBoxLayout()
        pattern_layout.addWidget(QLabel("Files with ending:"))
        pattern_widget = LineEdit(value=".tif")
        pattern_widget.changed.connect(self.setExpPattern)
        pattern_layout.addWidget(pattern_widget._widget._qwidget)

        main = QVBoxLayout()
        main.addLayout(pattern_layout)
        main.addLayout(DT_layout)
        main.addLayout(pixel_layout)

        return main

    ###
    # ACTION
    ###

    def load_pipeline(self, file_path):
        try:
            tp = ImagePipeline.from_yaml(file_path)
            self.tp = tp
        except BaseException as e:
            print(e)
            em = QErrorMessage()
            em.showMessage(
                "Could not load pipeline from file %s, check the format"
                % file_path
            )

    def run_pipeline(self):
        self.running = True
        self.enable_run_button()
        # self._run_pipeline()
        th = Thread(target=self._run_pipeline)
        th.start()

    def _run_pipeline(self):
        exp = self.exp
        assert exp is not None
        self.nFilesChanged.emit(len(exp.scan_folder()))
        self.progressChanged.emit(0)
        self.progressRunning.emit(True)
        for i, f in enumerate(exp):
            self.nameProcessingFile.emit("Processing %s ..." % f)
            acq = Acquisition(f, experiment=exp, image_pipeline=self.tp)
            if i == 0:
                pipeline_export_path_for_exp = self.tp.exp_params_path(acq)
                logging.info(
                    "Saving pipeline to %s" % pipeline_export_path_for_exp
                )
            self.tp.process(to_process=acq, force_reprocess=True)
            self.progressChanged.emit(i + 1)
        self.progressRunning.emit(False)
        self.enable_run_button()

    ###
    # PROPERTIES
    ###

    @property
    def tp(self):
        if hasattr(self, "_tp"):
            return self._tp
        else:
            return None

    @tp.setter
    def tp(self, tp: ImagePipeline):
        self._tp = tp
        self.enable_run_button()
        self.setWindowTitle("Palmari | %s" % tp.name)

    @property
    def exp(self):
        if hasattr(self, "_input_folder") and hasattr(self, "_export_folder"):
            exp = Experiment(
                data_folder=self.input_folder,
                export_folder=self.export_folder,
                DT=self.DT,
                file_pattern=self.file_pattern,
                pixel_size=self.pixel_size,
            )
            return exp
        else:
            return None

    @property
    def export_folder(self):
        if hasattr(self, "_export_folder"):
            return self._export_folder
        else:
            return None

    @export_folder.setter
    def export_folder(self, f: str):
        if os.path.exists(f) and os.path.isdir(f):
            self._export_folder = str(f)
            self.enable_run_button()

    def setExportFolder(self, f: str):
        self.export_folder = f

    @property
    def input_folder(self):
        if hasattr(self, "_input_folder"):
            return self._input_folder
        else:
            return None

    @input_folder.setter
    def input_folder(self, f: str):
        if os.path.exists(f) and os.path.isdir(f):
            self._input_folder = str(f)
            self.enable_run_button()

    def setInputFolder(self, f: str):
        self.input_folder = f

    ###
    # OTHERS
    ###

    def enable_run_button(self):
        exp = self.exp
        if self.running:
            self.run_button.setText("Processing...")
        else:
            if exp is not None:
                self.run_button.setText(
                    "Process %d files !" % len(exp.scan_folder())
                )
            else:
                self.run_button.setText("Process (?) files !")

        self.run_button.setEnabled(
            (self.tp is not None)
            and (self.export_folder is not None)
            and (self.input_folder is not None)
            and (not self.running)
        )

    def setExpPixel(self, s: float):
        self.pixel_size = s

    def setExpPattern(self, p: str):
        self.file_pattern = p

    def setExpDT(self, DT: float):
        self.DT = DT
